@@ -5,6 +5,9 @@ use strict;
 use warnings;
 use Carp;
 use Errno;
+use IO::Select;
+use IO::Socket;
+use Socket;
 
 =head1 NAME
 
@@ -12,7 +15,7 @@ Pick::TCL - class to run commands in a Pick TCL shell
 
 =head1 VERSION
 
-Version 0.04
+Version 0.05
 
 =cut
 
@@ -20,7 +23,7 @@ Version 0.04
 # PACKAGE GLOBALS #
 ###################
 
-our $VERSION = '0.04';
+our $VERSION = '0.05';
 our %_mods;
 
 #########################
@@ -56,6 +59,11 @@ if (scalar(keys %_mods) == 0)
     my $output = $ap->exec('TCL.COMMAND PARAMS (OPTIONS)');
     my $unsanitised = $ap->execraw('TCL.COMMAND PARAMS (OPTIONS)');
 
+    # Spawn, check & retrieve output of long-running commands
+    $ap->spawn('TCL.COMMAND PARAMS (OPTIONS)')} = 1;
+    # ...
+    do_something_with($ap->output) if $ap->is_ready;
+
     # Clean up
     $ap->logout();
 
@@ -63,7 +71,8 @@ if (scalar(keys %_mods) == 0)
 
 C<Pick::TCL> provides a class to run arbitrary B<TCL> (that's
 I<Terminal Control Language>, not the "other" TCL) commands in a
-local or remote Pick or Pick-like environment,
+local or remote Pick or Pick-like environment, either synchronously
+(blocking until execution has completed) or asynchronously.
 
 =over 4
 
@@ -148,6 +157,34 @@ sub _reconnect_ssh
     return $self->_bring_up_ssh();
 }
 
+# Generate user logon sequence
+sub _mklogon
+{
+    my $self = shift;
+    my $logon = '';
+    foreach my $k (qw/USER PASS MD MDPASS/)
+    {
+        next unless defined($$self{'_OPTIONS'}->{$k});
+        $logon .= $$self{'_OPTIONS'}->{$k} . "\n";
+    }
+    return $logon;
+}
+
+# Build Pick command
+sub _mkpickcmd
+{
+    my $self = shift;
+    my $tclcmd = shift;
+    my @args;
+    push @args, $$self{'_OPTIONS'}->{'PICKBIN'};
+    push @args, $$self{'_OPTIONS'}->{'OPTVM'};
+    push @args, $$self{'_OPTIONS'}->{'VM'};
+    push @args, $$self{'_OPTIONS'}->{'OPTSILENT'};
+    push @args, $$self{'_OPTIONS'}->{'OPTDATA'};
+    push @args, $self->_mklogon . $tclcmd . "\r";
+    return @args;
+}
+
 =head1 CLASS METHODS
 
 =head2 new(%options)
@@ -307,7 +344,7 @@ sub new
 =head2 $ap->exec($tclcmd, @input)
 
 Executes the Pick B<TCL> command C<$tclcmd> on the Pick VM associated
-with the C<Pick::TCL> object C<$ap> and returns the output.
+with the C<Pick::TCL> object C<$ap> synchronously and returns the output.
 
 In order to cope with the wide variety of terminal settings found on
 different Pick systems in the wild (or in some cases, even on different
@@ -357,24 +394,11 @@ sub execraw
     my $tclcmd = shift;
     my $input = undef;
     $input = join "\r", @_ if scalar(@_) > 0;
-
-    # User logon sequence
-    my $logon = "";
-    foreach my $k (qw/USER PASS MD MDPASS/)
-    {
-        $logon .= $$self{'_OPTIONS'}->{$k} . "\n"
-            if defined($$self{'_OPTIONS'}->{$k});
-    }
-
-    # Build Pick command
-    my @args = ( $$self{'_OPTIONS'}->{'PICKBIN'},
-        $$self{'_OPTIONS'}->{'OPTVM'}, $$self{'_OPTIONS'}->{'VM'},
-        $$self{'_OPTIONS'}->{'OPTSILENT'}, $$self{'_OPTIONS'}->{'OPTDATA'},
-        $logon . $tclcmd . "\r" );
-    my ($result, $err) = ("", "");
-    $ENV{'PATH'} = "";
+    my @args = $self->_mkpickcmd($tclcmd);
 
     # Run command
+    my ($result, $err) = ("", "");
+    $ENV{'PATH'} = "";
     if (defined($$self{'_SSH'}))
     {
         # Remote VM
@@ -396,7 +420,7 @@ sub execraw
         }
     } else {
         # Local VM
-        IPC::Run::run([ \@args, \$input, \$result, \$err ])
+        IPC::Run::run(\@args, \$input, \$result, \$err)
             or croak "Broken pipe to Pick: $!";
         if ($err)
         {
@@ -408,7 +432,6 @@ sub execraw
     return $result;
 }
 
-
 sub exec
 {
     my $self = shift;
@@ -427,7 +450,261 @@ sub exec
     return wantarray ? @lines : join "\n", @lines;
 }
 
-=head2 $ap->logout()
+=head2 $ap->spawn($tclcmd, @input)
+
+Spawns C<$tclcmd> for asynchronous execution on the Pick VM associated
+with C<$ap>.
+
+On success, returns true. On failure, returns C<undef> and sets C<$!>.
+
+Output can be retrieved later with the C<output> method (see below).
+
+=cut
+
+sub spawn
+{
+    my $self = shift;
+    my $func = 'Pick::TCL::spawnraw()';
+    croak "$func is not a class method" unless ref($self);
+    if (scalar(@_) == 0)
+    {
+        carp "$func: cowardly refusing to execute a null TCL command";
+        $! = &Errno::ENOMSG;
+        return undef;
+    }
+    my $tclcmd = shift;
+    my $input = undef;
+    $input = join "\r", @_ if scalar(@_) > 0;
+    my @args = $self->_mkpickcmd($tclcmd);
+
+    # Run command
+    $ENV{'PATH'} = "";
+    if (defined($$self{'_SSH'}))
+    {
+        # Remote VM
+        unless ($self->_reconnect_ssh())
+        {
+            carp "$func: $!";
+            return undef;
+        }
+        my $ssh = $$self{'_SSH'};
+        my ($socket, $pid) = $$ssh->open2socket({}, @args);
+        my $serr = $$ssh->error();
+        if ($serr)
+        {
+            carp "$func: fatal error: $serr: $! $?";
+            $! = &Errno::EBADFD;
+            return undef;
+        }
+        if (defined($input))
+        {
+            while (length($input) > 0)
+            {
+                my $bytes = syswrite $socket, $input;
+                unless (defined($bytes))
+                {
+                    carp "$func: socket write error: $!";
+                    return undef;
+                }
+                $input = susbtr($input, $bytes);
+            }
+        }
+        $socket->blocking(0);
+        my $set = IO::Select->new();
+        $set->add($socket);
+        $$self{'_SOCKET'} = $socket;
+        $$self{'_SOCKSET'} = $set;
+        $$self{'_PID'} = $pid;
+    } else {
+        # Local VM -- IPC::Run cannot detect EOF when spawning asynchronously,
+        #             so just fork() & run() instead
+        my ($child, $parent);
+        unless (socketpair($child, $parent, AF_UNIX, SOCK_STREAM, PF_UNSPEC))
+        {
+            carp "$func: socketpair(): $!";
+            return undef;
+        }
+        $child->autoflush(1);
+        $parent->autoflush(1);
+
+        my $pid = fork();
+        unless (defined($pid))
+        {
+            my $errno = $!;
+            carp "$func: fork(): $errno";
+            $child->close;
+            $parent->close;
+            $! = $errno;
+            return undef;
+        }
+        if ($pid)
+        {
+            # In parent, act just like we would for a remote VM
+            $parent->close;
+            my $set = IO::Select->new();
+            $set->add($child);
+            $$self{'_SOCKET'} = $child;
+            $$self{'_SOCKSET'} = $set;
+            $$self{'_PID'} = $pid;
+            return 1;
+        }
+        # In child, act just like we would for synchronous execution
+        $child->close;
+        my ($result, $err) = ('', '');
+        my $ran = IPC::Run::run(\@args, \$input, \$result, \$err)
+          or croak "Broken pipe to Pick: $!";
+        croak "$func: $err" if $err;
+        print $parent $result;
+        exit 0;
+    }
+    return 1;
+}
+
+=head2 $ap->is_ready
+
+Returns true if a previously spawned job has completed. Returns false
+but defined if the job is still running. Returns C<undef> if the job
+has aborted or if no job was spawned in the first place.
+
+=cut
+
+sub is_ready
+{
+    my $self = shift;
+    my $func = 'Pick::TCL::is_ready';
+    croak "$func is not a class method" unless ref($self);
+
+    # Same approach for local or remote VM
+    # -- both use sockets for asyncrhonous execution
+    unless (defined($$self{'_SOCKSET'}))
+    {
+        $! = &Errno::ENOENT;
+        return undef;
+    }
+    my $buf = '';
+    my $errflag = 0;
+    my $errno = 0;
+    while(1)
+    {
+        my $readable = scalar($$self{'_SOCKSET'}->can_read(0));
+        unless (defined($readable))
+        {
+            return 0 unless defined($$self{'_PARTIAL'});
+            $errflag++;
+            $errno = $!;
+            last;
+        }
+        $$self{'_PARTIAL'} = '' unless defined($$self{'_PARTIAL'});
+        last if $readable == 0;
+        my $res = sysread $$self{'_SOCKET'}, $buf, 65536, length($buf);
+        unless (defined($res))
+        {
+            carp "$func: socket error: $!" unless $! == &Errno::EAGAIN;
+            $errno = $!;
+            $errflag++;
+            last;
+        }
+        last if $res == 0;
+    }
+    $$self{'_PARTIAL'} .= $buf;
+    if (($errflag) && ($errno == 0))
+    {
+        return 0;
+    }
+    $$self{'_SOCKSET'}->remove($$self{'_SOCKET'});
+    $$self{'_SOCKET'}->close;
+    waitpid($$self{'_PID'}, 0);
+    delete $$self{'_SOCKSET'};
+    delete $$self{'_SOCKET'};
+    delete $$self{'_PID'};
+    if ($errflag)
+    {
+      $! = $errno;
+      return undef;
+    }
+    return 1;
+}
+
+=head2 $ap->output
+
+=head2 $ap->outputraw
+
+Returns the (raw or cooked) output of a previously spawned job if
+it has completed.
+
+Sets C<$!> and returns C<undef> (or, for the cooked form, the empty
+list if called in list context) if no job has been spawned.
+
+=cut
+
+sub outputraw
+{
+    my $self = shift;
+    my $func = 'Pick::TCL::outputraw';
+    croak "$func is not a class method" unless ref($self);
+    $$self{'_PARTIAL'} = '' unless defined($$self{'_PARTIAL'});
+
+    # Same approach for local or remote VM
+    # -- both use sockets for asyncrhonous execution
+    $! = &Errno::ENOENT unless defined($$self{'_PARTIAL'});
+    return $$self{'_PARTIAL'};
+}
+
+sub output
+{
+    my $self = shift;
+    my $func = 'Pick::TCL::output()';
+    croak "$func is not a class method" unless ref($self);
+
+    # Get response & sanitise
+    my $raw = $self->outputraw;
+    my $errno = $!;
+    if (defined($raw))
+    {
+        my @lines = split /[^\x09\x20-\x7e]+/m, $raw;
+        return wantarray ? @lines : join "\n", @lines;
+    }
+    $! = $errno;
+    return wantarray ? () : undef;
+}
+
+=head2 $ap->partialoutput
+
+=head2 $ap->partialoutputraw
+
+Returns the (raw or cooked) output of a previously spawned job
+that may not yet have completed.
+
+Returns C<undef> (or, for the cooked form, the empty list if
+called in list context) if a previously spawned job does not
+have any output yet or no job has been spawned.
+
+=cut
+
+sub partialoutputraw
+{
+    my $self = shift;
+    my $func = 'Pick::TCL::partialoutputraw';
+    croak "$func is not a class method" unless ref($self);
+    return $$self{'_PARTIAL'};
+}
+
+sub partialoutput
+{
+    my $self = shift;
+    my $func = 'Pick::TCL::partialoutput';
+    croak "$func is not a class method" unless ref($self);
+    my $raw = $self->partialoutputraw;
+    if (defined($raw))
+    {
+        my @lines = split /[^\x09\x20-\x7e]+/m, $raw;
+        return wantarray ? @lines : join "\n", @lines;
+    }
+    return wantarray ? () : undef;
+}
+ 
+
+=head2 $ap->logout
 
 Destroys the connection. Not required to be called explicitly before
 exit; does nothing when Pick is local.
@@ -466,6 +743,14 @@ to B<TCL> commands must be balanced (even though the Pick B<TCL>
 interpreter does not normally require that), as unbalanced parentheses
 will likely confuse the remote shell.
 
+=head2 Parallel asynchronous commands
+
+Each C<Pick::TCL> object can only spawn() or exec() one B<TCL>
+command at a time. If the caller needs to spawn() multiple B<TCL>
+commands in parallel (or exec() a B<TCL> command while waiting
+for an asynchronous one to complete), it should instantiate a
+separate C<Pick::TCL> object for each B<TCL> command.
+
 =head2 Pick flavours
 
 C<Pick::TCL> has only been tested with D3/Linux as the target Pick
@@ -495,7 +780,7 @@ Thanks to Jemalong Wool who provided funding for initial development.
 
 =head1 LICENSE AND COPYRIGHT
 
-Copyright 2013 Jack Burton.
+Copyright 2013, 2014 Jack Burton.
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of either: the GNU General Public License as published
